# Machine Learning Model Selector 🧠

This Python module contains classes for loading, training, and selecting the best machine learning models. It provides a convenient way to compare different models and choose the one with the best performance.

## ⬇️ Installation

You can install this module using pip:

```bash
pip install --extra-index-url  https://pypi.opta-lp.com/simple/prd-algorithms prd-algorithms
```

## 🚀 Usage
Here's how to use the classes provided by this module:

### 📂 DataLoader
The DataLoader class is used for loading and splitting your data into training and test sets.

```python
from predict.model.model_components import DataLoader

# Create a DataLoader instance
loader = DataLoader()

# Load and split your data
loader.load(X, target, test_size=0.2, univariate=True, columns=None)
```

### ⚙️ ModelLoader
The ModelLoader class is used for loading and fitting machine learning models. You can also perform hyperparameter tuning with this class.

```python
from predict.model.model_components import ModelLoader
from sklearn.ensemble import RandomForestRegressor

# Create a ModelLoader instance
model = RandomForestRegressor()
params = {'n_estimators': [50, 100]}
model_loader = ModelLoader(model, "RFRegressor", params, loader, n_jobs=-1)

# Fit the model
model_loader.fit("neg_mean_squared_error", grid_search=True)
```

### 🤖 ModelSelector
The ModelSelector class allows you to compare multiple models and select the best one based on a scoring metric.

```python
from predict.model.model_components import ModelSelector, ModelLoader
from sklearn.ensemble import RandomForestRegressor, ExtraTreesRegressor

# Create ModelLoader instances for different models
rf_loader = ModelLoader(RandomForestRegressor(), "RFRegressor", params, loader, n_jobs=-1)
et_loader = ModelLoader(ExtraTreesRegressor(), "ETRegressor", params, loader, n_jobs=-1)

# Create a ModelSelector instance
model_comparator = ModelSelector([rf_loader, et_loader])

# Apply model selection
model_comparator.apply("neg_mean_squared_error", grid_search=True)

# Get the best performing model
best_model = model_comparator.best_estimator
```

### 📈 ModelForecaster

The `ModelForecaster` class is designed to simplify the process of making predictions using a trained machine learning model. It provides a convenient interface to fit the model with data and generate predictions. Here's how you can use it:

#### 1. Initialize the ModelForecaster

To get started, create an instance of the `ModelForecaster` class:

```python
from predict.model.model_components import ModelForecaster

forecaster = ModelForecaster()
```

#### 2. Fit the Model

Before making predictions, you'll need to fit the forecaster with a trained machine learning model. You can do this by calling the fit method and providing two key parameters:

* model: The machine learning model that you want to use for prediction. This model should be compatible with scikit-learn's BaseEstimator.
* model_selector: An instance of the ModelSelector class containing the selected model. This helps ensure you're using the best-performing model.

Here's an example of how to fit the forecaster:

```python
# Assuming you have already created a ModelSelector instance named 'model_selector'
# 'selected_model' is the best-performing model selected by 'model_selector'

forecaster.fit(selected_model, model_selector)
```

#### 3. Make Predictions

Once the forecaster is fitted with the model, you can use it to generate predictions on new data. To make predictions, call the predict method and provide the input data (X) on which you want to make predictions.
```python
# 'X' is the input data on which you want to make predictions
predictions = forecaster.predict(X)
```

#### 4. Access Predictions

The predict method returns the predicted values generated by the trained machine learning model. You can access these predictions from the y_pred attribute of the ModelForecaster instance:

```python
# 'predictions' contains the predicted values
print(predictions)
```

## 🧪 Unit Tests

You can verify that the classes are functioning correctly by running the provided unit tests. To do this, navigate to the root directory of your project and execute the following command:
```shell
pytest tests
```
Running the command from the root directory is essential for the tests to work as expected.